GIT 0.99.9j aka 1.0rc3
[git.git] / git-am.sh
index c562159..98a390a 100755 (executable)
--- a/git-am.sh
+++ b/git-am.sh
@@ -3,16 +3,10 @@
 #
 . git-sh-setup || die "Not a git archive"
 
-files=$(git-diff-index --cached --name-only HEAD) || exit
-if [ "$files" ]; then
-   echo "Dirty index: cannot apply patches (dirty: $files)" >&2
-   exit 1
-fi
-
 usage () {
     echo >&2 "usage: $0 [--signoff] [--dotest=<dir>] [--utf8] [--3way] <mbox>"
     echo >&2 " or, when resuming"
-    echo >&2 " $0 [--skip]"
+    echo >&2 " $0 [--skip | --resolved]"
     exit 1;
 }
 
@@ -35,10 +29,10 @@ fall_back_3way () {
     mkdir "$dotest/patch-merge-tmp-dir"
 
     # First see if the patch records the index info that we can use.
-    if git-apply --show-index-info "$dotest/patch" \
+    if git-apply -z --index-info "$dotest/patch" \
        >"$dotest/patch-merge-index-info" 2>/dev/null &&
        GIT_INDEX_FILE="$dotest/patch-merge-tmp-index" \
-       git-update-index --index-info <"$dotest/patch-merge-index-info" &&
+       git-update-index -z --index-info <"$dotest/patch-merge-index-info" &&
        GIT_INDEX_FILE="$dotest/patch-merge-tmp-index" \
        git-write-tree >"$dotest/patch-merge-base+" &&
        # index has the base tree now.
@@ -104,7 +98,7 @@ fall_back_3way () {
 }
 
 prec=4
-dotest=.dotest sign= utf8= keep= skip= interactive=
+dotest=.dotest sign= utf8= keep= skip= interactive= resolved=
 
 while case "$#" in 0) break;; esac
 do
@@ -128,6 +122,9 @@ do
        -k|--k|--ke|--kee|--keep)
        keep=t; shift ;;
 
+       -r|--r|--re|--res|--reso|--resol|--resolv|--resolve|--resolved)
+       resolved=t; shift ;;
+
        --sk|--ski|--skip)
        skip=t; shift ;;
 
@@ -140,6 +137,8 @@ do
        esac
 done
 
+# If the dotest directory exists, but we have finished applying all the
+# patches in them, clear it out.
 if test -d "$dotest" &&
    last=$(cat "$dotest/last") &&
    next=$(cat "$dotest/next") &&
@@ -153,10 +152,11 @@ if test -d "$dotest"
 then
        test ",$#," = ",0," ||
        die "previous dotest directory $dotest still exists but mbox given."
+       resume=yes
 else
-       # Make sure we are not given --skip
-       test ",$skip," = ,, ||
-       die "we are not resuming."
+       # Make sure we are not given --skip nor --resolved
+       test ",$skip,$resolved," = ,,, ||
+               die "we are not resuming."
 
        # Start afresh.
        mkdir -p "$dotest" || exit
@@ -169,12 +169,24 @@ else
                exit 1
        }
 
+       # -s, -u and -k flags are kept for the resuming session after
+       # a patch failure.
+       # -3 and -i can and must be given when resuming.
        echo "$sign" >"$dotest/sign"
        echo "$utf8" >"$dotest/utf8"
        echo "$keep" >"$dotest/keep"
        echo 1 >"$dotest/next"
 fi
 
+case "$resolved" in
+'')
+       files=$(git-diff-index --cached --name-only HEAD) || exit
+       if [ "$files" ]; then
+          echo "Dirty index: cannot apply patches (dirty: $files)" >&2
+          exit 1
+       fi
+esac
+
 if test "$(cat "$dotest/utf8")" = t
 then
        utf8=-u
@@ -215,31 +227,50 @@ do
                go_next
                continue
        }
-       git-mailinfo $keep $utf8 "$dotest/msg" "$dotest/patch" \
-               <"$dotest/$msgnum" >"$dotest/info" ||
-               stop_here $this
-       git-stripspace < "$dotest/msg" > "$dotest/msg-clean"
+
+       # If we are not resuming, parse and extract the patch information
+       # into separate files:
+       #  - info records the authorship and title
+       #  - msg is the rest of commit log message
+       #  - patch is the patch body.
+       #
+       # When we are resuming, these files are either already prepared
+       # by the user, or the user can tell us to do so by --resolved flag.
+       case "$resume" in
+       '')
+               git-mailinfo $keep $utf8 "$dotest/msg" "$dotest/patch" \
+                       <"$dotest/$msgnum" >"$dotest/info" ||
+                       stop_here $this
+               git-stripspace < "$dotest/msg" > "$dotest/msg-clean"
+               ;;
+       esac
 
        GIT_AUTHOR_NAME="$(sed -n '/^Author/ s/Author: //p' "$dotest/info")"
        GIT_AUTHOR_EMAIL="$(sed -n '/^Email/ s/Email: //p' "$dotest/info")"
        GIT_AUTHOR_DATE="$(sed -n '/^Date/ s/Date: //p' "$dotest/info")"
-       SUBJECT="$(sed -n '/^Subject/ s/Subject: //p' "$dotest/info")"
+       export GIT_AUTHOR_NAME GIT_AUTHOR_EMAIL GIT_AUTHOR_DATE
 
+       SUBJECT="$(sed -n '/^Subject/ s/Subject: //p' "$dotest/info")"
        case "$keep_subject" in -k)  SUBJECT="[PATCH] $SUBJECT" ;; esac
-       if test '' != "$SIGNOFF"
-       then
+
+       case "$resume" in
+       '')
+           if test '' != "$SIGNOFF"
+           then
                LAST_SIGNED_OFF_BY=`
-                       sed -ne '/^Signed-off-by: /p' "$dotest/msg-clean" |
-                       tail -n 1
+                   sed -ne '/^Signed-off-by: /p' \
+                   "$dotest/msg-clean" |
+                   tail -n 1
                `
-               ADD_SIGNOFF=$(test "$LAST_SIGNED_OFF_BY" = "$SIGNOFF" || {
+               ADD_SIGNOFF=`
+                   test "$LAST_SIGNED_OFF_BY" = "$SIGNOFF" || {
                    test '' = "$LAST_SIGNED_OFF_BY" && echo
                    echo "$SIGNOFF"
-               })
-       else
+               }`
+           else
                ADD_SIGNOFF=
-       fi
-       {
+           fi
+           {
                echo "$SUBJECT"
                if test -s "$dotest/msg-clean"
                then
@@ -250,8 +281,18 @@ do
                then
                        echo "$ADD_SIGNOFF"
                fi
-       } >"$dotest/final-commit"
+           } >"$dotest/final-commit"
+           ;;
+       *)
+               case "$resolved,$interactive" in
+               tt)
+                       # This is used only for interactive view option.
+                       git-diff-index -p --cached HEAD >"$dotest/patch"
+                       ;;
+               esac
+       esac
 
+       resume=
        if test "$interactive" = t
        then
            test -t 0 ||
@@ -263,14 +304,17 @@ do
                echo "--------------------------"
                cat "$dotest/final-commit"
                echo "--------------------------"
-               echo -n "Apply? [y]es/[n]o/[e]dit/[a]ccept all "
+               echo -n "Apply? [y]es/[n]o/[e]dit/[v]iew patch/[a]ccept all "
                read reply
                case "$reply" in
-               y*|Y*) action=yes ;;
-               a*|A*) action=yes interactive= ;;
-               n*|N*) action=skip ;;
-               e*|E*) "${VISUAL:-${EDITOR:-vi}}" "$dotest/final-commit"
+               [yY]*) action=yes ;;
+               [aA]*) action=yes interactive= ;;
+               [nN]*) action=skip ;;
+               [eE]*) "${VISUAL:-${EDITOR:-vi}}" "$dotest/final-commit"
                       action=again ;;
+               [vV]*) action=again
+                      LESS=-S ${PAGER:-less} "$dotest/patch" ;;
+               *)     action=again ;;
                esac
            done
        else
@@ -293,7 +337,21 @@ do
        echo "Applying '$SUBJECT'"
        echo
 
-       git-apply --index "$dotest/patch"; apply_status=$?
+       case "$resolved" in
+       '')
+               git-apply --index "$dotest/patch"
+               apply_status=$?
+               ;;
+       t)
+               # Resolved means the user did all the hard work, and
+               # we do not have to do any patch application.  Just
+               # trust what the user has in the index file and the
+               # working tree.
+               resolved=
+               apply_status=0
+               ;;
+       esac
+
        if test $apply_status = 1 && test "$threeway" = t
        then
                if (fall_back_3way)